extern crate pest;
#[macro_use]
extern crate pest_derive;

mod actions;
mod types;
mod utils;

use actions::FileStack;
use colored::*;
use pest::error::ErrorVariant::{CustomError, ParsingError};
use pest::error::LineColLocation;
use pest::iterators::Pair;
use pest::iterators::Pairs;
use pest::{Parser, Span};
use std::io;
use std::io::Read;
use std::path::PathBuf;

#[derive(Parser)]
#[grammar = "grammar.pest"]
pub struct IviaParser;

#[derive(Debug)]
pub struct Command<'a> {
    pub name: String,
    pub params: Vec<Param<'a>>,
    pub at: Span<'a>,
}

#[derive(Debug)]
pub struct Param<'a> {
    pub name: &'a str,
    pub value: Pair<'a, Rule>,
}

fn main() {
    let (input_files, contents, output_files) =
        into_paths_and_program(std::env::args().skip(1).collect());

    let mut stack = match FileStack::new(input_files) {
        Ok(result) => result,
        Err(error_message) => {
            println!("{}", error_message.red());
            return;
        }
    };

    println!("{:?}", contents);

    let commands = match IviaParser::parse(Rule::program, &contents.trim()) {
        Ok(pairs) => into_commands(pairs),
        Err(e) => {
            show_parsing_error(e);
            vec![]
        }
    };

    // println!("{:#?}", commands);
    pretty(&commands);

    match stack.do_commands(&commands) {
        Err(e) => {
            println!("{}", e.message.red());
            println!("{:?}", e.at);
        }
        Ok(()) => {
            println!("{}", "Done".green());
        }
    }

    // let stack = actions::FileStack::new(input_files)?;
}

fn show_parsing_error(error: pest::error::Error<Rule>) {
    match error.line_col {
        LineColLocation::Pos((_, col)) => {
            print!("{}^", " ".to_owned().repeat(col + 1 - 1));
        }
        LineColLocation::Span((_, start), (_, end)) => {
            print!("{}^", " ".to_owned().repeat(start + 1 - 1));
            print!("{}", "~".to_owned().repeat(end - start));
        }
    }
    match error.variant {
        ParsingError {
            positives,
            negatives: _,
        } => {
            println!(" expected {:?}", positives);
        }
        CustomError { message } => {
            println!("{}", message);
        }
    }
}

fn pretty(commands: &Vec<Command>) {
    for command in commands {
        println!("· {name}", name = command.name.blue());

        fn pretty_rule(pair: &Pair<Rule>, indent: usize, name: String) {
            let pair_value = pair.as_str();
            println!(
                "{}{}{} {} {}",
                "  ".repeat(indent),
                if name != "" {
                    format!("({}) ", name.italic()).dimmed()
                } else {
                    "".to_owned().bold()
                },
                format!("{:?}", pair.as_rule()).green(),
                if pair_value.len() < 20 {
                    pair_value
                } else {
                    ""
                },
                format!("@ {}~{}", pair.as_span().start(), pair.as_span().end()).dimmed()
            );
            for inner_pair in pair.clone().into_inner() {
                pretty_rule(&inner_pair, indent + 1, "".to_owned());
            }
        }
        for param in &command.params {
            pretty_rule(&param.value, 2, param.name.to_string());
        }
    }
}

fn into_commands<'a>(pairs: Pairs<'a, Rule>) -> Vec<Command<'a>> {
    let mut commands: Vec<Command<'a>> = Vec::new();
    for pipeline in pairs.into_iter() {
        for command in pipeline.into_inner() {
            let mut current_command: Command<'a> = Command {
                name: String::new(),
                params: Vec::new(),
                at: command.as_span(),
            };
            for pair in command.into_inner() {
                match pair.as_rule() {
                    Rule::action => current_command.name = pair.as_str().to_string(),
                    Rule::param => {
                        let mut current_param_name = "";
                        for param_thing in pair.into_inner() {
                            match param_thing.as_rule() {
                                Rule::param_name => {
                                    current_param_name = param_thing.as_str()
                                }
                                Rule::param_value => {
                                    for value in param_thing.into_inner() {
                                        current_command.params.push(Param {
                                            name: current_param_name,
                                            value,
                                        })
                                    }
                                }
                                _ => (),
                            };
                        }
                    }
                    Rule::param_value => {
                        for value in pair.into_inner() {
                            current_command.params.push(Param {
                                name: "",
                                value,
                            });
                        }
                    }
                    _ => (),
                }
            }
            commands.push(current_command);
        }
    }
    commands
}

fn into_paths_and_program(cli_args: Vec<String>) -> (Vec<PathBuf>, String, Vec<PathBuf>) {
    let mut contents = String::new();
    let mut input_files: Vec<PathBuf> = Vec::new();
    let mut output_files: Vec<PathBuf> = Vec::new();
    if cli_args.len() != 0 {
        let mut current_cli_arg_is_input = true;
        let last_argument_index = cli_args.len() - 1;
        cli_args.into_iter().enumerate().for_each(|(i, arg)| {
            let as_path = PathBuf::from(&arg);
            if i == last_argument_index {
                output_files.push(as_path);
                return;
            }
            if current_cli_arg_is_input {
                match IviaParser::parse(Rule::action, &arg) {
                    Ok(_) => {
                        current_cli_arg_is_input = false;
                        contents.push_str(&arg);
                        contents.push_str(" ");
                    }
                    Err(_) => {
                        input_files.push(as_path);
                    }
                }
            } else {
                contents.push_str(&arg);
                contents.push_str(" ")
            }
        });
    }
    if contents.as_str() == "" {
        match io::stdin().read_to_string(&mut contents) {
            Ok(_) => (),
            Err(e) => {
                eprintln!("Could not read line from stdin: {}", e);
                std::process::exit(1);
            }
        }
    };
    (input_files, contents.trim().to_owned(), output_files)
}
